import torch
import torch.nn as nn
import numpy as np
import random
import torch.nn.utils.rnn as rnn_utils
from transformer.Models import Encoder

def dynamic_rnn(inputs, lens, rnn):
    
    def enc(inputs, inputs_len, rnn):
        
        #实现Variable lengths输入RNN
        packed = rnn_utils.pack_padded_sequence(inputs, lengths=inputs_len.tolist(), batch_first=True)
        packed_out, states = rnn(packed)
        unpacked_, _ = rnn_utils.pad_packed_sequence(packed_out, batch_first=True)
        return unpacked_, states

    def sort_by_len(inputs, lens):
        
        inputs_len, sort_ids = torch.sort(lens, 0, descending=True)
        inputs = inputs[sort_ids]
        _, origin_ids = torch.sort(sort_ids, 0, descending=False)
        
        return inputs, inputs_len, origin_ids
    
    inputs, lens_, origin_ids = sort_by_len(inputs, lens)
    outputs, states = enc(inputs, lens_, rnn)
    outputs = outputs[origin_ids]
    if 'LSTM' in str(rnn.type):
        states = (states[0].transpose(0,1)[origin_ids], states[1].transpose(0,1)[origin_ids])
    else:
        states = states.transpose(0,1)[origin_ids] # batch, num_layers*num_directions, hidden_size
        
    return outputs, states
    

class RNNEncoder(nn.Module):
    
    def __init__(self, CONFIG):
        
        super(RNNEncoder, self).__init__()
        self.CONFIG = CONFIG
        
        # self.maxpooling1d = nn.MaxPool1d(CONFIG.max_sent_len, stride=2)
        # 变长pooling，并且可以k-pooling
        self.maxpooling1d = nn.AdaptiveMaxPool1d(1)
        self.dropout = nn.Dropout(CONFIG.dropout)        
        self.set_rnn()
        self.set_activation()
    
    def set_activation(self):
        
        if self.CONFIG.activation == 'ReLU':
            self.activation = nn.ReLU()
        if self.CONFIG.activation == 'sigmoid':
            self.activation = nn.Sigmoid()
        if self.CONFIG.activation == 'tanh':
            self.activation = nn.Tanh()

    def set_rnn(self):
        
        if self.CONFIG.enc == 'GRU':

            self.enc = nn.GRU(input_size=self.CONFIG.embedding_dim,
                              hidden_size=self.CONFIG.hidden_size,
                              num_layers=self.CONFIG.rnn_layers,
                              dropout=self.CONFIG.dropout,
                              bidirectional=self.CONFIG.bidirectional,
                              batch_first=True)

        elif self.CONFIG.enc == 'LSTM':

            self.enc = nn.LSTM(input_size=self.CONFIG.embedding_dim,
                               hidden_size=self.CONFIG.hidden_size,
                               num_layers=self.CONFIG.rnn_layers,
                               dropout=self.CONFIG.dropout,
                               bidirectional=self.CONFIG.bidirectional,
                               batch_first=True)

            
    def forward(self, inputs, lens):
        
        outputs, states = dynamic_rnn(inputs, lens, self.enc)
        '''取最后一个hidden state'''
        if self.CONFIG.pooling_type == 'last':
            fw_, bw_ = torch.split(outputs, self.CONFIG.hidden_size, -1)
            size_ = fw_.size()
            
            '''根据seq_len获取正向hidden state'''
            fw_ = fw_.contiguous().view(size_[0]*size_[1], -1)
            offset = torch.Tensor([i*size_[1]-1 for i in range(size_[0])]).cuda()
            index_ = lens.long() + offset.long()
            fw_vec = torch.index_select(fw_, 0, index_.long())
            '''后向rnn，取第一个hidden state'''
            bw_vec = bw_[:, 0, :]
            
            sentence_vector = torch.cat([fw_vec, bw_vec], 1)
        
        '''maxpooling 得到sentence vector'''
        if self.CONFIG.pooling_type == 'maxpooling':
            outputs = torch.transpose(outputs, 1, 2)
            sentence_vector = self.maxpooling1d(outputs)
        
        sentence_vector = sentence_vector.squeeze(-1)
        
        return sentence_vector
        #return self.dropout(sentence_vector)

        

class CNNEncoder(nn.Module):
    
    def __init__(self, CONFIG):
        
        super(CNNEncoder, self).__init__()
        self.CONFIG = CONFIG
        
        # self.maxpooling1d = nn.MaxPool1d(CONFIG.max_sent_len, stride=2)
        # 变长pooling，并且可以k-pooling
        self.maxpooling1d = nn.AdaptiveMaxPool1d(1)
        self.dropout = nn.Dropout(CONFIG.dropout)        
        self.set_cnn()
        self.set_activation()

    def set_activation(self):

        if self.CONFIG.activation == 'ReLU':
            self.activation = nn.ReLU()
        if self.CONFIG.activation == 'sigmoid':               
            self.activation = nn.Sigmoid()
        if self.CONFIG.activation == 'tanh':
            self.activation = nn.Tanh()

    def set_cnn(self):
        
        cnns = []
        poolings = []
        for i in range(self.CONFIG.cnn_layers):
            if i == 0:
                cnns.append(nn.Conv1d(in_channels=self.CONFIG.embedding_dim,
                                      out_channels=self.CONFIG.channels,
                                      kernel_size=self.CONFIG.kernel_size,
                                      dilation=self.CONFIG.dilations[0]))
            else:
                cnns.append(nn.Conv1d(in_channels=self.CONFIG.channels,
                                      out_channels=self.CONFIG.channels,
                                      kernel_size=self.CONFIG.kernel_size,
                                      dilation=self.CONFIG.dilations[i]))
            if self.CONFIG.pooling_type == 'maxpooling':
                poolings = nn.AdaptiveMaxPool1d(1)
            elif self.CONFIG.pooling_type == 'meanpooling':
                poolings = nn.AdaptiveAvgPool1d(1)
        
        self.cnns = nn.ModuleList(cnns)
        self.poolings = poolings
        
            
    def forward(self, inputs, lens):
        
        inputs = torch.transpose(inputs, 1, 2)
        for i in range(len(self.cnns)):
            inputs = self.cnns[i](inputs)
        
        sentence_vector = self.poolings(inputs).squeeze(-1)
        sentence_vector = self.activation(sentence_vector)
        return sentence_vector
        
        
        
class TransformerEncoder(nn.Module):
    
    def __init__(self, CONFIG, src_word_emb):
        
        super(TransformerEncoder, self).__init__()
        self.CONFIG = CONFIG
        self.enc = Encoder(src_word_emb=src_word_emb,
                          len_max_seq=CONFIG.max_sent_len,
                          d_word_vec=CONFIG.embedding_dim,
                          n_layers=CONFIG.n_layers,
                          n_head=CONFIG.n_head,
                          d_k=CONFIG.d_k,
                          d_v=CONFIG.d_v,
                          d_model=CONFIG.transformer_hidden_size,
                          d_inner=CONFIG.d_inner,
                          dropout=CONFIG.dropout)
        self.maxpooling1d = nn.AdaptiveMaxPool1d(1)
        self.dropout = nn.Dropout(CONFIG.dropout) 
        
    
    def forward(self, src_seq, src_pos, lens, return_attns=False):
        
        outputs = self.enc(src_seq, src_pos, return_attns)
        #outputs = self.maxpooling1d(outputs[0].transpose(1, 2)).squeeze(-1)
        outputs = outputs[0][:, 0, :]
        return outputs
        