# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

from __future__ import absolute_import, division, unicode_literals

"""
Example of file for SkipThought in SentEval
"""
import logging
import sys
import torch
import json
import argparse
import os


# Set PATHs
PATH_TO_SENTEVAL = '../'
PATH_TO_DATA = '../data/'

parser = argparse.ArgumentParser()
parser.add_argument("-mp", "--model_path", help="pretrained model path")
parser.add_argument("-ev", "--expand_vector", help="expand vector")
parser.add_argument("-voc", "--voc_path", help="expand vocab")
parser.add_argument("-e", "--expand_embedding", type=int, default=0, help="whether to expand vocab")
parser.add_argument("-gpu", "--gpu",  default = '0', help="which gpu to run")

args = parser.parse_args()

# import skipthought and Senteval
sys.path.insert(0, PATH_TO_SENTEVAL)
import senteval
from net import SeqEncoder, Encoder
from config import Config


bookcorpus_dir = '../../../data/bookcorpus/'
lm_dir = '../../../data/1B/1-billion-word-language-modeling-benchmark-r13output/'


def load_voc():

    return json.load(open(args.voc_path))

def prepare(params, samples):
    
    total_words, nuk_words  = set(), set()
    for s in samples:
        for word in s:
            total_words.add(word.lower())
            if word.lower() not in params['voc']:
                nuk_words.add(word)

    logging.info('Found {0} nuk words with word vectors, out of \
        {1} words'.format(len(nuk_words), len(total_words)))
 
    return


def load_model(model_path):
   
    net = SeqEncoder()
    checkpoint = torch.load(args.model_path)
    if 'net' in checkpoint:
        net.load_state_dict(checkpoint['net'])
    else:
        net.load_state_dict(checkpoint)
    '''expand vocab'''
    if args.expand_embedding:
        net.load_pretrained_embedding(args.expand_vector)
    net = net.to(device)
    net.eval()
    return net

def batcher(params, batch):

    voc = params_senteval['voc']
    batch = [['-<begin>-'] + sent for sent in batch]
    batch = [sent if sent != [] else '.' for sent in batch]
    inputs = []
    lengths = []
    maxlen = -1
    for i in range(len(batch)):
        maxlen = len(batch[i]) if maxlen < len(batch[i]) else maxlen

    for i in range(len(batch)):
        tmp = [voc['-<eos>-'] for j in range(maxlen)]
        len_ = 0
        for j in range(min(len(batch[i]), maxlen)):
            if batch[i][j].lower() in voc:
                tmp[j] = voc[batch[i][j].lower()]
            else:
                tmp[j] = voc['-<unk>-']
            len_ += 1
        inputs.append(tmp)
        lengths.append(len_)

    inputs = torch.LongTensor(inputs).to(params_senteval['device'])
    lengths = torch.LongTensor(lengths).to(params_senteval['device'])
    #  embeddings = params['encoder'].seq_enc(inputs, lengths)
    #  sum_ = torch.sum(embeddings*embeddings, 1).view(-1, 1)
    #  embeddings = (embeddings/sum_).detach().cpu().numpy()
    embeddings = params['encoder'].seq_enc(inputs, lengths).detach().cpu().numpy()
    return embeddings

# Set params for SentEval
params_senteval = {'task_path': PATH_TO_DATA, 'usepytorch': True, 'kfold': 10, 'batch_size':128}
params_senteval['classifier'] = {'nhid': 0, 'optim': 'adam', 'batch_size': 128, 'usepytorch':True, 
        'tenacity': 3, 'epoch_size': 4, 'cudaEfficient': True}

# Set up logger
logging.basicConfig(format='%(asctime)s : %(message)s', level=logging.DEBUG)

if __name__ == "__main__":
    # Load SkipThought model

    gpus = [_ for _ in args.gpu.split('-')]
    os.environ["CUDA_VISIBLE_DEVICES"] = ','.join(gpus)
   
    device = torch.device('cuda')
    params_senteval['encoder'] = load_model(args.model_path)
    params_senteval['device'] = device
    params_senteval['voc'] = load_voc()
    
    transfer_tasks = ['STS12', 'STS13', 'STS14', 'STS15', 'STS16',
                      'MR', 'CR', 'MPQA', 'SUBJ', 'SST2', 'SST5', 'TREC', 'MRPC',
                      'SICKEntailment', 'SICKRelatedness', 'STSBenchmark',
                      'Length', 'WordContent', 'Depth', 'TopConstituents',
                      'BigramShift', 'Tense', 'SubjNumber', 'ObjNumber',
                      'OddManOut', 'CoordinationInversion']

    bzDict = json.load(open('bz.json'))
    for task in transfer_tasks:
        
        bz = bzDict[task] if task in bzDict else 128
        params_senteval['batch_size'] = bz
        params_senteval['seed'] = 2048
        params_senteval['classifier'] = {'nhid': 0, 'optim': 'adam', 'batch_size': bz, 'usepytorch':True, 
                'tenacity': 3, 'epoch_size': 4, 'cudaEfficient': True}

        se = senteval.engine.SE(params_senteval, batcher, prepare)
        results = se.eval([task])
    
