import torch
import torch.nn as nn
import numpy as np
import torch.nn.utils.rnn as rnn_utils
from enc import *
 
class SeqEncoder(nn.Module):
    
    def __init__(self, CONFIG):
        
        super(SeqEncoder, self).__init__()
        
        self.CONFIG = CONFIG
        self.dropout = nn.Dropout(CONFIG.dropout)
        self.BCELoss = nn.BCEWithLogitsLoss()
        self.RankingLoss = nn.MarginRankingLoss(CONFIG.ranking_margin)
        self.CosineDis = nn.CosineSimilarity()
        self.init_embedding()
        self.init_enc(CONFIG)
            
        # sentvec 和 word 映射到一个空间
        self.sentvec_mapping_word = nn.Linear(in_features=self.sentvec_size, out_features=CONFIG.mapping_dim)
        self.word_mapping_sentvec = nn.Linear(in_features=CONFIG.embedding_dim, out_features=CONFIG.mapping_dim)
        
        # 判断sent是不是一个语序正常的句子
        self.order_fc = nn.Linear(in_features=self.sentvec_size, out_features=1)
        
        if CONFIG.normfunc == 'xariver':
            self.xariver_init()
        
    def init_enc(self, CONFIG):
        
        if CONFIG.enc == 'GRU' or CONFIG.enc == 'LSTM':
            self.enc = RNNEncoder(CONFIG)
            self.sentvec_size = 2 if CONFIG.bidirectional else 1
            self.sentvec_size *= CONFIG.hidden_size
        if CONFIG.enc == 'CNN':
            self.enc = CNNEncoder(CONFIG)
            self.sentvec_size = CONFIG.channels
        if CONFIG.enc == 'transformer':
            self.enc = TransformerEncoder(CONFIG, self.embedding)
            self.sentvec_size = CONFIG.transformer_hidden_size
        if CONFIG.enc == 'CGRU' or CONFIG.enc == 'CLSTM':
            self.enc = CRNNEncoder(CONFIG)
            self.sentvec_size = 2 if CONFIG.bidirectional else 1
            self.sentvec_size *= CONFIG.hidden_size
        if CONFIG.enc == 'BiFlowGRU' or CONFIG.enc == 'BiFlowLSTM':
            self.enc = BiFlow(CONFIG)
            self.sentvec_size = CONFIG.hidden_size
        if CONFIG.enc == 'StackLSTM' or CONFIG.enc == 'StackGRU':
            self.enc = StackRNNEncoder(CONFIG)
            self.sentvec_size = CONFIG.hidden_size * self.CONFIG.rnn_layers * 2
    
    
    def init_embedding(self):
        
        if self.CONFIG.use_glove:
            self.load_pretrained_embedding(self.CONFIG.vec_path)
        else:
            self.embedding = nn.Embedding(num_embeddings=self.CONFIG.voc_size, embedding_dim=self.CONFIG.embedding_dim)
            nn.init.uniform_(self.embedding.weight, -0.1, 0.1)
            
    def xariver_init(self):
        
        for key in self.state_dict():
            if key.split('.')[-1] == 'weight':
                if not 'embedding' in key:
                    try:
                        nn.init.xavier_normal_(self.state_dict()[key])
                    except:
                        print ('can not xariver init ', key)
            elif key.split('.')[-1] == 'bias':
                if 'rnn' in key.split('.')[-1]:
                    self.state_dict()[key] = 1
                    
            
    def load_pretrained_embedding(self, embedding):
        
        ''' 加载预训练向量'''
        pretrained_weight = torch.from_numpy(np.load(embedding)).float()
        self.embedding = nn.Embedding.from_pretrained(pretrained_weight, freeze=True)


    def ranking_forward(self, sentvecs):

        '''
        相邻句子的余弦距离比非相邻句子的余弦距离大
        '''
        query, x1, x2 = torch.split(sentvecs, 1, 1)
        query, x1, x2 = query.squeeze(1), x1.squeeze(1), x2.squeeze(1)
        
        qx1_dis = self.CosineDis(query, x1)
        qx2_dis = self.CosineDis(query, x2)
        
        predictions = qx1_dis-qx2_dis
        predictions[predictions > 0] = 1
        predictions[predictions <= 0] = -1
        
        
        return (qx1_dis, qx2_dis), predictions.int().cpu().data.numpy()

        
    def word_forward(self, word_inputs, sentence_vector):
       
        '''
        词向量与句子向量的点乘，正例点乘大，负例点乘小
        '''
        word_inputs = self.embedding(word_inputs)
        
        expand_vector = sentence_vector.view(sentence_vector.size(0), 1, sentence_vector.size(1))
        expand_vector = expand_vector.expand(sentence_vector.size(0), word_inputs.size(1), sentence_vector.size(1))
        sent_ = self.sentvec_mapping_word(expand_vector)
        word_ = self.word_mapping_sentvec(word_inputs)
        dot = sent_ * word_
        logits = torch.sum(dot, -1).squeeze(-1)
        
        return logits, logits.cpu().detach().numpy() > 0
        
    def order_forward(self, sentence_vector):
        
        '''
        判断一个句子词序是否正常
        '''
        logits = self.order_fc(sentence_vector).squeeze(-1)
        
        return logits, logits.cpu().detach().numpy() > 0
        
    def word_loss(self, word_logits, word_label):
        
        return self.BCELoss(word_logits, word_label.float())
    
    def order_loss(self, order_logits, order_label):
        
        return self.BCELoss(order_logits, order_label.float())

    def ranking_loss(self, x, label):
            
        return self.RankingLoss(x[0], x[1], label.float())

    def seq_enc(self, inputs, lens):
       
        if self.CONFIG.enc == 'transformer':
            pos = [[i for i in range(inputs.size(1))] for j in range(inputs.size(0))]
            pos = torch.LongTensor(pos).to(inputs.device)
            return self.enc(inputs, pos, lens)
        else:
            inputs = self.embedding(inputs)
            sentvec = self.enc(inputs, lens)
            return sentvec
        
    def forward(self): 
        return 

