import torch
import torch.nn as nn
import numpy as np
import random
import torch.nn.utils.rnn as rnn_utils
from transformer.Models import Encoder

def dynamic_rnn(inputs, lens, rnn):
    
    '''
    动态RNN函数，实现封装编码解封都操作
    '''
    def enc(inputs, inputs_len, rnn):
        
        #实现Variable lengths输入RNN
        packed = rnn_utils.pack_padded_sequence(inputs, lengths=inputs_len.tolist(), batch_first=True)
        packed_out, states = rnn(packed)
        unpacked_, _ = rnn_utils.pad_packed_sequence(packed_out, batch_first=True)
        return unpacked_, states

    def sort_by_len(inputs, lens):
        
        inputs_len, sort_ids = torch.sort(lens, 0, descending=True)
        inputs = inputs[sort_ids]
        _, origin_ids = torch.sort(sort_ids, 0, descending=False)
        
        return inputs, inputs_len, origin_ids
    
    inputs, lens_, origin_ids = sort_by_len(inputs, lens)
    outputs, states = enc(inputs, lens_, rnn)
    outputs = outputs[origin_ids]
    if 'LSTM' in str(rnn.type):
        states = (states[0].transpose(0,1)[origin_ids], states[1].transpose(0,1)[origin_ids])
    else:
        states = states.transpose(0,1)[origin_ids] # batch, num_layers*num_directions, hidden_size
        
    return outputs, states

def bi_pooling(inputs, pooling_type, lens):
        
    '''
    对双向RNN编码结果进行pooling得到输出，分别有maxpooling和lastpooling两种
    maxpooling：直接对输出沿着时间维度取最大值
    lastpooling：根据输入原长度与batch_size，利用index_select进行取值
    '''
    
    '''取最后一个hidden state'''
    if pooling_type == 'last':
        fw_, bw_ = torch.split(inputs, hidden_size, -1)
        size_ = fw_.size()

        '''根据seq_len获取正向hidden state'''
        fw_ = fw_.contiguous().view(size_[0]*size_[1], -1)
        offset = torch.Tensor([i*size_[1]-1 for i in range(size_[0])]).cuda()
        index_ = lens.long() + offset.long()
        fw_vec = torch.index_select(fw_, 0, index_.long())
        '''后向rnn，取第一个hidden state'''
        bw_vec = bw_[:, 0, :]

        sentence_vector = torch.cat([fw_vec, bw_vec], 1)

    '''maxpooling 得到sentence vector'''
    if pooling_type == 'maxpooling':
        inputs = torch.transpose(inputs, 1, 2)
        sentence_vector = nn.functional.adaptive_max_pool1d(inputs, 1)

    sentence_vector = sentence_vector.squeeze(-1)
    return sentence_vector

def uni_pooling(inputs, pooling_type, lens):
    
    '''
    对单向RNN编码结果进行pooling得到输出，分别有maxpooling和lastpooling两种
    maxpooling：直接对输出沿着时间维度取最大值
    lastpooling：根据输入原长度与batch_size，利用index_select进行取值
    '''
    
    '''取最后一个hidden state'''
    if pooling_type == 'last':
        
        size_ = inputs.size()

        '''根据seq_len获取正向hidden state'''
        inputs = inputs.contiguous().view(size_[0]*size_[1], -1)
        offset = torch.Tensor([i*size_[1]-1 for i in range(size_[0])]).cuda()
        index_ = lens.long() + offset.long()
        sentence_vector = torch.index_select(inputs, 0, index_.long())
    
    '''maxpooling 得到sentence vector'''
    if pooling_type == 'maxpooling':
        inputs = torch.transpose(inputs, 1, 2)
        sentence_vector = nn.functional.adaptive_max_pool1d(inputs, 1)

    sentence_vector = sentence_vector.squeeze(-1)
    return sentence_vector

def pooling(inputs, pooling_type, lens, bidirectional=True):
    
    if bidirectional:
        return bi_pooling(inputs, pooling_type, lens)
    else:
        return uni_pooling(inputs, pooling_type, lens)
       

class RNNEncoder(nn.Module):
   
    '''
    RNN 编码器
    '''
    def __init__(self, CONFIG):
        
        super(RNNEncoder, self).__init__()
        self.CONFIG = CONFIG
        
        # self.maxpooling1d = nn.MaxPool1d(CONFIG.max_sent_len, stride=2)
        # 变长pooling，并且可以k-pooling
        self.maxpooling1d = nn.AdaptiveMaxPool1d(1)
        self.dropout = nn.Dropout(CONFIG.dropout)        
        self.set_rnn()
        self.set_activation()
    
    def set_activation(self):
        
        if self.CONFIG.activation == 'ReLU':
            self.activation = nn.ReLU()
        if self.CONFIG.activation == 'sigmoid':
            self.activation = nn.Sigmoid()
        if self.CONFIG.activation == 'tanh':
            self.activation = nn.Tanh()

    def set_rnn(self):
        
        if self.CONFIG.enc == 'GRU':
            func = eval('nn.GRU')
        if self.CONFIG.enc == 'LSTM':
            func = eval('nn.LSTM')
        self.rnn = func(input_size=self.CONFIG.embedding_dim,
                        hidden_size=self.CONFIG.hidden_size,
                        num_layers=self.CONFIG.rnn_layers,
                        dropout=self.CONFIG.dropout,
                        bidirectional=self.CONFIG.bidirectional,
                        batch_first=True)
      
    def forward(self, inputs, lens):
        
        inputs = self.dropout(inputs)
        outputs, states = dynamic_rnn(inputs, lens, self.rnn)
        sentence_vector = pooling(outputs, self.CONFIG.pooling_type, lens, bidirectional=self.CONFIG.bidirectional)
        return self.dropout(sentence_vector)

        

class CNNEncoder(nn.Module):
    
    '''
    多层CNN编码器
    '''
    def __init__(self, CONFIG):
        
        super(CNNEncoder, self).__init__()
        self.CONFIG = CONFIG
        
        # self.maxpooling1d = nn.MaxPool1d(CONFIG.max_sent_len, stride=2)
        # 变长pooling，并且可以k-pooling
        self.maxpooling1d = nn.AdaptiveMaxPool1d(1)
        self.dropout = nn.Dropout(CONFIG.dropout)        
        self.set_cnn()
        self.set_activation()

    def set_activation(self):

        if self.CONFIG.activation == 'ReLU':
            self.activation = nn.ReLU()
        if self.CONFIG.activation == 'sigmoid':               
            self.activation = nn.Sigmoid()
        if self.CONFIG.activation == 'tanh':
            self.activation = nn.Tanh()

    def set_cnn(self):
        
        cnns = []
        poolings = []
        bns = []
        for i in range(self.CONFIG.cnn_layers):
            if i == 0:
                cnns.append(nn.Conv1d(in_channels=self.CONFIG.embedding_dim,
                                      out_channels=self.CONFIG.channels,
                                      kernel_size=self.CONFIG.kernel_size,
                                      dilation=self.CONFIG.dilations[0]))
                bns.append(nn.BatchNorm1d(self.CONFIG.channels))
            else:
                cnns.append(nn.Conv1d(in_channels=self.CONFIG.channels,
                                      out_channels=self.CONFIG.channels,
                                      kernel_size=self.CONFIG.kernel_size,
                                      dilation=self.CONFIG.dilations[i]))
                bns.append(nn.BatchNorm1d(self.CONFIG.channels))
            if self.CONFIG.pooling_type == 'maxpooling':
                poolings = nn.AdaptiveMaxPool1d(1)
            elif self.CONFIG.pooling_type == 'meanpooling':
                poolings = nn.AdaptiveAvgPool1d(1)
        
        self.cnns = nn.ModuleList(cnns)
        self.bns = nn.ModuleList(bns)
        self.poolings = poolings
        
            
    def forward(self, inputs, lens):
        
        inputs = self.dropout(inputs)
        inputs = torch.transpose(inputs, 1, 2)
        for i in range(len(self.cnns)):
            inputs = self.cnns[i](inputs)
            #inputs = self.bns[i](inputs)
            inputs = self.dropout(inputs) # dropout
            inputs = self.activation(inputs)
        
        sentence_vector = self.poolings(inputs).squeeze(-1)
        return sentence_vector # 已经有dropout了
        

class CRNNEncoder(nn.Module):
   
    '''
    CRNN编码器，对输入先进行CNN编码，再进行RNN编码
    '''
    def __init__(self, CONFIG):
        
        super(CRNNEncoder, self).__init__()
        self.CONFIG = CONFIG
        self.set_cnn()
        self.set_rnn()
        self.set_activation()
        self.dropout = nn.Dropout(CONFIG.dropout)
     
    def set_activation(self):

        if self.CONFIG.activation == 'ReLU':
            self.activation = nn.ReLU()
        if self.CONFIG.activation == 'sigmoid':               
            self.activation = nn.Sigmoid()
        if self.CONFIG.activation == 'tanh':
            self.activation = nn.Tanh()

    def set_cnn(self):
        
        cnns = []
        for i, kernel_size in enumerate(self.CONFIG.crnn_kernel_sizes):
            cnns.append(nn.Conv1d(in_channels=self.CONFIG.embedding_dim,
                                  out_channels=self.CONFIG.crnn_channels[i],
                                  kernel_size=kernel_size))
        self.cnns = nn.ModuleList(cnns)
    
    def set_rnn(self):
        
        embedding_dim = sum(self.CONFIG.crnn_channels) + self.CONFIG.embedding_dim
        if self.CONFIG.enc == 'CGRU':
            func = eval('nn.GRU')
        if self.CONFIG.enc == 'CLSTM':
            func = eval('nn.LSTM')
        self.rnn = func(input_size=embedding_dim,
                        hidden_size=self.CONFIG.hidden_size,
                        num_layers=self.CONFIG.rnn_layers,
                        dropout=self.CONFIG.dropout,
                        bidirectional=self.CONFIG.bidirectional,
                        batch_first=True)

    def padding(self, inputs, padding_=(1,1)):
        
        size = inputs.size()
        padding_var = torch.zeros((size[0], size[1])).unsqueeze(2).to(inputs.device)
        if padding_[0]:
            tmp = padding_var.expand(size[0], size[1], padding_[0])
            inputs = torch.cat([tmp, inputs], -1)
        if padding_[1]:
            tmp = padding_var.expand(size[0], size[1], padding_[1])
            inputs = torch.cat([inputs, tmp], -1)
        return inputs
        
    def cnn_enc(self, inputs):
        
        inputs = inputs.transpose(1, 2)
        outputs = []
        for i, kernel_size in enumerate(self.CONFIG.crnn_kernel_sizes):
            if kernel_size % 2:
                inputs_ = self.padding(inputs, ((kernel_size-1)//2, (kernel_size-1)//2))
                outputs_ = self.cnns[i](inputs_).transpose(1, 2)
                outputs.append(outputs_)
            else:
                inputs_ = self.padding(inputs, (1+(kernel_size-2)//2, (kernel_size-1)//2))
                outputs_ = self.cnns[i](inputs_).transpose(1, 2)
                outputs.append(outputs_)
            
        outputs = torch.cat(outputs, -1)
        return self.dropout(self.activation(outputs))
        
    def forward(self, inputs, lens):
        
        inputs = self.dropout(inputs)
        outputs = self.cnn_enc(inputs)
        inputs = torch.cat([outputs, inputs], -1)
        outputs, states = dynamic_rnn(inputs, lens, self.rnn)
        sentence_vector = pooling(outputs, self.CONFIG.pooling_type, lens, bidirectional=self.CONFIG.bidirectional)
        return self.dropout(sentence_vector)
        
class BiFlow(nn.Module):
   
    '''
    双向RNN，先正向对输入进行编码，再对编码输出进行逆序再一次编码
    '''
    def __init__(self, CONFIG):
        
        super(BiFlow, self).__init__()
        self.CONFIG = CONFIG
        self.set_rnn()
        self.set_activation()
        self.dropout = nn.Dropout(CONFIG.dropout)
     
    def set_activation(self):

        if self.CONFIG.activation == 'ReLU':
            self.activation = nn.ReLU()
        if self.CONFIG.activation == 'sigmoid':               
            self.activation = nn.Sigmoid()
        if self.CONFIG.activation == 'tanh':
            self.activation = nn.Tanh()

    
    def set_rnn(self):
        
        if self.CONFIG.enc == 'BiFlowGRU':
            func = eval('nn.GRU')
        if self.CONFIG.enc == 'BiFlowLSTM':
            func = eval('nn.LSTM')
        
        self.bw_rnn = func(input_size=self.CONFIG.embedding_dim,
                           hidden_size=self.CONFIG.hidden_size,
                           batch_first=True)
        self.fw_rnn = func(input_size=self.CONFIG.hidden_size,
                           hidden_size=self.CONFIG.hidden_size,
                           batch_first=True)
        
    def forward(self, inputs, lens):
        
        inputs = self.dropout(inputs)
        idx = torch.LongTensor([i for i in range(inputs.size(1)-1, -1, -1)]).to(inputs.device)
        reverse_inputs = inputs.index_select(1, idx)
        bw_hiddens, _ = self.bw_rnn(reverse_inputs)
        bw_hiddens = bw_hiddens.index_select(1, idx) # 再次翻转得到原本的顺序
        bw_hiddens = self.dropout(bw_hiddens)
        
        outputs, states = dynamic_rnn(bw_hiddens, lens, self.fw_rnn)
        sentence_vector = pooling(outputs, self.CONFIG.pooling_type, lens, bidirectional=self.CONFIG.bidirectional)
        return self.dropout(sentence_vector)
        

class StackRNNEncoder(nn.Module):
    
    '''
    多层RNN编码，并且将每一层的编码输出进行拼接
    '''
    def __init__(self, CONFIG):
        
        super(StackRNNEncoder, self).__init__()
        
        self.CONFIG = CONFIG
        self.set_rnn()
        self.set_activation()
        self.dropout = nn.Dropout(CONFIG.dropout)
     
    def set_activation(self):

        if self.CONFIG.activation == 'ReLU':
            self.activation = nn.ReLU()
        if self.CONFIG.activation == 'sigmoid':               
            self.activation = nn.Sigmoid()
        if self.CONFIG.activation == 'tanh':
            self.activation = nn.Tanh()

    
    def set_rnn(self):
        
        if self.CONFIG.enc == 'StackGRU':
            func = eval('nn.GRU')
        if self.CONFIG.enc == 'StackLSTM':
            func = eval('nn.LSTM')
        
        fws = []
        bws = []
        fws.append(func(input_size=self.CONFIG.embedding_dim,
                        hidden_size=self.CONFIG.hidden_size,
                        batch_first=True))
        bws.append(func(input_size=self.CONFIG.embedding_dim,
                        hidden_size=self.CONFIG.hidden_size,
                        batch_first=True))
        for i in range(1, self.CONFIG.rnn_layers):
            fws.append(func(input_size=self.CONFIG.hidden_size,
                            hidden_size=self.CONFIG.hidden_size,
                            batch_first=True))
            bws.append(func(input_size=self.CONFIG.hidden_size,
                            hidden_size=self.CONFIG.hidden_size,
                            batch_first=True))
        self.fws = nn.ModuleList(fws)
        self.bws = nn.ModuleList(bws)
        
            
    
    def forward(self, inputs, lens):
        
        inputs = self.dropout(inputs)
        fws = []
        bws = []
        fw_hiddens, _ = dynamic_rnn(inputs, lens, self.fws[0])
        fws.append(fw_hiddens)
        idx = torch.LongTensor([i for i in range(inputs.size(1)-1, -1, -1)]).to(inputs.device)
        reverse_inputs = inputs.index_select(1, idx)
        bw_hiddens, _ = self.bws[0](reverse_inputs)
        bws.append(bw_hiddens)
        
        for i in range(1, len(self.fws)):
            inputs = self.dropout(fws[-1])
            fw_hiddens, _ = dynamic_rnn(inputs, lens, self.fws[i])
            fws.append(fw_hiddens)
            
            inputs = self.dropout(bws[-1])
            bw_hiddens, _ = self.bws[i](inputs)
            bws.append(bw_hiddens)
            
        fws = torch.cat(fws, -1)
        bws = torch.cat(bws, -1).index_select(1, idx) # 恢复原本顺序
        sentence_vector = torch.cat([fws, bws], -1)
        sentence_vector = pooling(sentence_vector, self.CONFIG.pooling_type, lens, bidirectional=False)
        return self.dropout(sentence_vector)
        
    
        
class TransformerEncoder(nn.Module):
    
    '''
    Transformer 编码
    '''
    def __init__(self, CONFIG, src_word_emb):
        
        super(TransformerEncoder, self).__init__()
        self.CONFIG = CONFIG
        self.enc = Encoder(src_word_emb=src_word_emb,
                          len_max_seq=CONFIG.max_sent_len,
                          d_word_vec=CONFIG.embedding_dim,
                          n_layers=CONFIG.n_layers,
                          n_head=CONFIG.n_head,
                          d_k=CONFIG.d_k,
                          d_v=CONFIG.d_v,
                          d_model=CONFIG.transformer_hidden_size,
                          d_inner=CONFIG.d_inner,
                          dropout=CONFIG.dropout)
        self.maxpooling1d = nn.AdaptiveMaxPool1d(1)
        self.dropout = nn.Dropout(CONFIG.dropout) 
        
    
    def forward(self, src_seq, src_pos, lens, return_attns=False):
        
        outputs, _ = self.enc(src_seq, src_pos, return_attns)
        outputs = outputs[:, 0, :]
        return outputs # 已经有dropout了
        
