import numpy as np
from collections import Counter
import h5py
import random
import json
import sys
from tqdm import *
import lmdb 

import os
from config import Config

from nltk.tokenize import sent_tokenize
from nltk import word_tokenize


def build_voc(input_list, voc_path, case_sentive=True):

    '''
    建立字典
    input_list：多个输入文件路径
    voc_path：字典存储路径，包括字典与单词计数
    case_sentive：是否大小写敏感
    '''
    finished_num = 0
    raw_voc = Counter()
    for input_path in input_list:
        print (input_path)
        with open(input_path) as inf:
            for ele in inf:
                ele = ele.rstrip() if case_sentive else ele.rstrip().lower()
                '''切分成句子，bookcorpus是单句，等同无效'''
                for _ in ele.split('<sep>'):
                    raw_voc.update(_.split())

                finished_num += 1
                if finished_num % 100000 == 0:
                    print ("%d finished" % finished_num)

    
    sorted_voc = sorted(raw_voc.items(), key=lambda item : item[1], reverse=True)
    word_nums = dict()
    voc = dict()
    voc['</s>'] = Config.eos_id
    voc['<s>'] = Config.bos_id
    voc['[unk]'] = Config.unk_id
    word_nums[voc['</s>']] = 1
    word_nums[voc['<s>']] = 1
    word_nums[voc['[unk]']] = 1
    
    for i, ele in enumerate(sorted_voc):
        if len(voc) >= Config.voc_size:
            break
        if ele[0] in voc:
            continue
        voc[ele[0]] = len(voc)
        word_nums[voc[ele[0]]] = ele[1]
            
    print ('build voc ready')
    
    json.dump(voc, open(voc_path + 'voc.json', 'w'), indent=2)
    json.dump(word_nums, open(voc_path + 'voc_num.json', 'w'), indent=2)


def build_voc_in_w2v(input_list, w2v_voc, w2v_vec, voc_path, case_sentive=True):
    
    '''
    建立预训练词向量字典，对语料字典中，只保存有预训练词向量的单词
    input_list：多个输入文件路径
    w2v_voc：预训练词向量字典
    w2v_vec：预训练词向量向量
    voc_path：字典存储路径，包括字典与单词计数
    case_sentive：是否大小写敏感
    '''
    
    w2v_voc = json.load(open(w2v_voc))
    w2v_vec = np.load(w2v_vec)
    
    raw_voc = Counter()
    for input_path in input_list:
        with open(input_path) as inf:
            finished_num = 0
            for ele in inf:
                ele = ele.rstrip() if case_sentive else ele.rstrip().lower()
                '''切分成句子，bookcorpus是单句，等同无效'''
                for _ in ele.split('<sep>'):
                    raw_voc.update(_.split())
                    
                finished_num += 1
                if finished_num % 100000 == 0:
                    print ("%d finished" % finished_num)

    
    word_nums = dict()
    voc = dict()
    voc['</s>'] = Config.eos_id
    voc['<s>'] = Config.bos_id
    voc['[unk]'] = Config.unk_id
    word_nums[voc['</s>']] = 1
    word_nums[voc['<s>']] = 1
    word_nums[voc['[unk]']] = 1
    
    for k, v in raw_voc.items():
        if k in w2v_voc and k not in voc:
            voc[k] = len(voc)
            word_nums[voc[k]] = v
    
    vec = [0 for i in range(len(voc))]
    for k,v in voc.items():
        if k in w2v_voc:
            vec[v] = w2v_vec[w2v_voc[k]]
        else:
            vec[v] = [0 for i in range(w2v_vec.shape[1])]
    
    print ('build voc ready')
    
    json.dump(voc, open(voc_path + 'w2v_voc.json', 'w'), indent=2)
    json.dump(word_nums, open(voc_path + 'w2v_voc_num.json', 'w'), indent=2)
    np.save(voc_path+'w2v_vec.npy', vec)
    
    
def tokenize(input_path, output_path):
    
    '''
    将输入文件进行分句分词，句子以<sep>分隔，词以空格分隔
    input_path：输入文件路径
    output_path：输出文件路径
    '''
    finished_num = 0
    ouf = open(output_path, 'w')
    with open(input_path) as inf:
        for text in inf:
            separated_text = ''
            sents = sent_tokenize(text.rstrip('\n'))
            for sent in sents:
                words = word_tokenize(sent)
                separated_text += ' '.join(words)
                separated_text += '<sep>'
            
            separated_text = separated_text.rstrip('<sep>')
            
            finished_num += 1
            ouf.write(separated_text+'\n')
            if  finished_num % 50000 == 0:
                print ('%d finished' % finished_num)
    
    ouf.close()

def txt2lmdb(input_path, output_path, type_):
    
    '''
    将输入文件进行分句分词，句子以<sep>分隔，词以空格分隔
    input_path：输入文件路径
    output_path：输出文件路径
    '''
    env = lmdb.open(output_path, map_size=1000000000099) 
    txn = env.begin(write=True) 
    finished_num = txn.stat()['entries']
    
    type_ = type_ + '-'
    finished_num = txn.stat()['entries']
    with open(input_path) as inf:
        for text in inf:
            txn.put(key=(type_+str(finished_num)).encode(), value=text.rstrip().encode())
            finished_num += 1
            
            if  finished_num % 100000 == 0:
                print ('%d finished' % finished_num)
                
    txn.commit() 
    env.close()
    
def t2i_lmdb(input_path, output_path, voc_path, type_, case_sentive):
    
    '''
    把text转化为index，并且以lmdb格式保存
    input_path：输入文件路径，已经提前分句分词，句子以<sep>分隔，词以空格分词
    output_path：保存路径
    voc_path：字典路径
    type_: 数据来源，bookcorpus、umbc等
    '''
    voc = json.load(open(voc_path))
    env = lmdb.open(output_path, map_size=1000000000099) 
    txn = env.begin(write=True) 

    type_ = type_ + '-'
    finished_num = txn.stat()['entries']
    with open(input_path) as inf:
        for text in inf:
            separated_text = ''
            sents = text if case_sentive else text.lower()
            sents = text.rstrip('\n').split('<sep>')
            for sent in sents:
                words = []
                for _ in sent.split():
                    words.append(str(voc.get(_, voc['[unk]'])))
                separated_text += ' '.join(words)
                separated_text += '<sep>'
            
            separated_text = separated_text.rstrip('<sep>')
            txn.put(key=(type_+str(finished_num)).encode(), value=separated_text.rstrip().encode())
            finished_num += 1
            
            if  finished_num % 100000 == 0:
                print ('%d finished' % finished_num)
                
    txn.commit() 
    env.close()
            
                
            

if __name__ == '__main__':
    
    '''input_file1, input_file2, output_voc_dir, case_sentive'''
    if sys.argv[1] == 'build_voc':
        build_voc([sys.argv[2], sys.argv[3]], sys.argv[4], int(sys.argv[5]))
    '''input_file1, input_file2, w2v_voc, w2v_vec, output_voc_dir, case_sentive'''
    if sys.argv[1] == 'build_voc_in_w2v':
        build_voc_in_w2v([sys.argv[2], sys.argv[3]], sys.argv[4], sys.argv[5], sys.argv[6], int(sys.argv[7]))
    '''input_path, output_path, voc_path, type, case_sentive'''
    if sys.argv[1] == 't2i_lmdb':
        t2i_lmdb(sys.argv[2], sys.argv[3], sys.argv[4], sys.argv[5], int(sys.argv[6]))
    '''input_path, output_path'''
    if sys.argv[1] == 'tokenize':
        tokenize(sys.argv[2], sys.argv[3])
    '''input_path, output_path, type'''
    if sys.argv[1] == 'txt2lmdb':
        txt2lmdb(sys.argv[2], sys.argv[3], sys.argv[4])
    
