import torch
import os, sys
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
import json
import random
import torch.nn.utils.rnn as rnn_utils
from torch.utils.data import DataLoader
from torch.nn import utils
import argparse
import logging
from torch.nn import DataParallel
import time

from folder import Folder
from folder import _collate_fn
from net import *
from enc import *
from config import Config
from logger import Logger
from utils import set_logger

parser = argparse.ArgumentParser()
parser.add_argument("-gpu", "--gpu",  default = '0', help="which gpu to run")
args = parser.parse_args()    


def save(net, optimizer, global_steps, trained_samples, prec, loss, keep_name=False):

    '''
    保存模型与优化器
    '''
    if not os.path.isdir(Config.model_dir):
        os.mkdir(Config.model_dir)
    
    if keep_name:
        path = Config.model_dir + 'prec-{:3f}-loss-{:3f}-steps-{:d}.pkl'.format(prec, loss, global_steps)
    else:
        path = Config.model_dir + 'best.pkl'

    state = {'net':net.state_dict(), 
             'optimizer':optimizer.state_dict(), 
             'global_steps':global_steps,
             'trained_samples':trained_samples}
    torch.save(state, path)
    
def pack_(data):
    
    '''
    把词预测、词排序、句子排序的句子进行打包，有利于并行化
    同时编码词预测、词排序与句子排序的句子，而不需要分成三次编码
    '''
    inputs = data['intra_input'].to(device)
    length = data['intra_len'].to(device)
    order_input = data['order_input'].to(device)
    ranking_input = data['ranking_input'].to(device)
    ranking_length = data['ranking_length'].to(device)
        
    input_list, input_lens, input_sizes, len_sizes = [], [], [], []
    
    if Config.word_coef:
        input_sizes.append(inputs.size())
        len_sizes.append(length.size())
        input_list.append(inputs)
        input_lens.append(length.view(-1))
        
    if Config.order_coef:
        input_sizes.append(order_input.size())
        len_sizes.append(length.size())
        input_list.append(order_input)
        input_lens.append(length.view(-1))
        
    if Config.ranking_coef:
        input_sizes.append(ranking_input.size())
        len_sizes.append(ranking_length.size())
        input_list.append(ranking_input.view(-1, ranking_input.size(-1)))
        input_lens.append(ranking_length.view(-1))
        
    return (torch.cat(input_list, 0), torch.cat(input_lens, 0).view(-1)), (input_sizes, len_sizes)
        
def unpack_(sentvecs, sizes):
    
    '''
    将编码结果进行拆解，根据距离的size
    '''
    input_sizes, len_sizes = sizes[0], sizes[1]
    bz, id_ = input_sizes[0][0], 0
    sentvec, ooo_sentvec, ranking_sentvec = 0, 0, 0
    
    if Config.word_coef:
        sentvec = sentvecs[:bz] # bz, sent_size
        id_ += 1
    
    if Config.order_coef:
        ooo_sentvec = sentvecs[id_*bz:(id_+1)*bz] # bz, sent_size
        id_ += 1
    
    if Config.ranking_coef:
        ranking_sentvec = sentvecs[id_*bz:] # bz*num_sents_for_one_ranking, sent_size
        ranking_sentvec = ranking_sentvec.view(input_sizes[id_][0], input_sizes[id_][1], -1)
    
    return sentvec, ooo_sentvec, ranking_sentvec  
    
def valid(valid_loader, net):

   
    num_word_samples, num_order_samples, num_ranking_samples = 0, 0, 0
    word_vprec, order_vprec, ranking_vprec, word_loss, order_loss, ranking_loss = 0, 0, 0, 0, 0, 0
    
    for (_, data) in enumerate(valid_loader, 0):
        
        if len(data) == 0:
            continue
        
        word_input = data['word_input'].to(device)
        word_label = data['word_label'].to(device)
        order_label = data['order_label'].to(device)
        ranking_label = data['ranking_label'].to(device)
        
        loss1, loss2, loss3, word_logits, order_logits, ranking_logits = .0, .0, .0, .0, .0, .0
        
        packed_inputs, sizes = pack_(data)
        packed_sentvec = net.seq_enc(packed_inputs[0], packed_inputs[1])
        sentvec, ooo_sentvec, ranking_sentvec = unpack_(packed_sentvec, sizes)
        
        if Config.word_coef:
            word_logits, word_pred = net.word_forward(word_input, sentvec)
            loss1 = net.word_loss(word_logits, word_label)
            word_loss += np.sum(loss1.cpu().data.numpy()) * word_label.size(0) * word_label.size(1)
            _vprec = np.sum(word_pred == word_label.cpu().data.numpy())
            word_vprec += _vprec
        
        if Config.order_coef:
            '''
            sentvec 就是正例，因此在这里构造对应size的正例标签
            '''
            sentvecs = torch.cat([sentvec, ooo_sentvec], 0)
            pos_label, neg_label = torch.ones_like(order_label), order_label
            order_label = torch.cat([pos_label, neg_label], 0).view(-1)
            order_logits, order_pred = net.order_forward(sentvecs)
            loss2 = net.order_loss(order_logits, order_label)
            order_loss += np.sum(loss2.cpu().data.numpy()) * order_label.size(0)
            _vprec = np.sum(order_pred == order_label.cpu().data.numpy()) 
            order_vprec += _vprec

        if Config.ranking_coef:
            ranking_logits, ranking_pred = net.ranking_forward(ranking_sentvec)
            loss3 = net.ranking_loss(ranking_logits, ranking_label)
            ranking_loss += np.sum(loss3.cpu().data.numpy()) * ranking_label.size(0)
            _vprec = np.sum(ranking_pred == ranking_label.cpu().data.numpy())
            ranking_vprec += _vprec

        num_word_samples += word_label.size(0) * word_label.size(1)
        num_order_samples += order_label.size(0)
        num_ranking_samples += ranking_label.size(0)

        del loss1, loss2, loss3, word_logits, order_logits, ranking_logits
    
    word_vprec /= num_word_samples
    word_loss /= num_word_samples
    order_vprec /= num_order_samples
    order_loss /= num_order_samples
    ranking_loss /= num_ranking_samples
    ranking_vprec /= num_ranking_samples
    
    return word_vprec, word_loss, order_vprec, order_loss, ranking_vprec, ranking_loss
        

def decay_lr(optimizer):
    
    '''
    学习率调整
    '''
    msg = ''
    for param_group in optimizer.param_groups:         
        param_group['lr'] = 0.99 * param_group['lr']
        msg = 'adjust lr %s' % param_group['lr']
    return msg

def train(train_loader, valid_loader, net, checkpoint):

    net = net.to(device)
    optimizer = torch.optim.Adam(filter(lambda p: p.requires_grad, net.parameters()), lr = Config.lr, weight_decay = Config.weight_decay)
    global_steps = 0
    trained_samples = 0
    if checkpoint:
        net.load_state_dict(checkpoint['net'])
        optimizer.load_state_dict(checkpoint['optimizer'])
        global_steps = checkpoint['global_steps']
        trained_samples = checkpoint['trained_samples']
    
    last_saved_step = global_steps
    last_saved_time = time.time()
    last_adjust_lr_step = global_steps
    best_loss = 10e7
    best_prec = -1
    tensorboard_tlogger = Logger(Config.tensorboard_tdir)
    tensorboard_vlogger = Logger(Config.tensorboard_vdir)
    logging = set_logger(Config.model_dir+'log.txt')

    for epoch in range(Config.epoch):
        '''超过一定的batch验证集不变好，就early stop'''
        #  if global_steps - last_saved_step > Config.early_stopping_batch:
        #      break
        for (_, data) in enumerate(train_loader, 0):
            
            if len(data) == 0:
                continue
            word_input = data['word_input'].to(device)
            word_label = data['word_label'].to(device)
            order_label = data['order_label'].to(device)
            ranking_label = data['ranking_label'].to(device)
                        
            word_loss, order_loss, ranking_loss, word_tprec, order_tprec, ranking_tprec = .0, .0, .0, .0, .0, .0

            packed_inputs, sizes = pack_(data)
            packed_sentvec = net.seq_enc(packed_inputs[0], packed_inputs[1])
            sentvec, ooo_sentvec, ranking_sentvec = unpack_(packed_sentvec, sizes)
        
            if Config.word_coef:
                word_logits, word_pred = net.word_forward(word_input, sentvec)
                word_loss = net.word_loss(word_logits, word_label)
                word_tprec = np.sum(word_pred == word_label.cpu().data.numpy()) / (word_label.size(0) * word_label.size(1))
                
            if Config.order_coef:
                
                sentvecs = torch.cat([sentvec, ooo_sentvec], 0)
                pos_label, neg_label = torch.ones_like(order_label), order_label
                order_label = torch.cat([pos_label, neg_label], 0).view(-1)
                order_logits, order_pred = net.order_forward(sentvecs)
                order_loss = net.order_loss(order_logits, order_label)
                order_tprec = np.sum(order_pred == order_label.cpu().data.numpy()) / order_label.size(0)

            if Config.ranking_coef:
                ranking_logits, ranking_pred = net.ranking_forward(ranking_sentvec)
                ranking_loss = net.ranking_loss(ranking_logits, ranking_label)
                ranking_tprec = np.sum(ranking_pred == ranking_label.cpu().data.numpy()) / ranking_label.size(0)

            global_steps += 1
            trained_samples += word_label.size(0)
            optimizer.zero_grad()
            loss = Config.word_coef * word_loss + Config.order_coef * order_loss + Config.ranking_coef * ranking_loss
            loss.backward()
            nn.utils.clip_grad_norm_(net.parameters(), 5)
            optimizer.step()

            
            if global_steps % Config.print_per_batch == 0:

                info = {}
                info['word_loss'] = word_loss
                info['word_prec'] = word_tprec
                info['order_loss'] = order_loss
                info['order_prec'] = order_tprec
                info['ranking_loss'] = ranking_loss
                info['ranking_prec'] = ranking_tprec
                info['total_loss'] = word_loss + order_loss + ranking_loss
                info['trained samples'] = trained_samples
                for tag, value in info.items():
                    tensorboard_tlogger.scalar_summary(tag, value, global_steps)
                msg = 'Train  iter: {0:>6}, epoch: {1:>2},  word loss: {2:>6.2}, word prec: {3:>6.2%}, order loss: {4:>6.2}, order prec: {5:>6.2%},  ranking loss: {6:>6.2},  ranking prec: {7:>6.2%}'
                logging.info(msg.format(global_steps, epoch, word_loss, word_tprec, order_loss, order_tprec, ranking_loss, ranking_tprec))
                
            #查看验证集当前情况
            if global_steps % Config.valid_per_batch == 0:

                net.eval()
                word_vprec, word_vloss, order_vprec, order_vloss, ranking_vprec, ranking_vloss =  valid(valid_loader, net)
                total_loss = word_vloss + order_vloss +  ranking_vloss
                total_prec = word_vprec + order_vprec +  ranking_vprec
                info = {}
                info['word_loss'] = word_vloss
                info['word_prec'] = word_vprec
                info['order_loss'] = order_vloss
                info['order_prec'] = order_vprec
                info['total_loss'] = total_loss
                info['ranking_loss'] = ranking_vloss
                info['ranking_prec'] = ranking_vprec
                info['trained samples'] = trained_samples
                for tag, value in info.items():
                    tensorboard_vlogger.scalar_summary(tag, value, global_steps)

                better = "" 
                net.train()
                if total_loss < best_loss:  # and total_prec > best_prec:
                    best_loss = total_loss
                    best_prec = total_prec 
                    last_saved_step = global_steps
                    save(global_steps = global_steps,
                        loss = best_loss,
                        prec = best_prec,
                        net=net,
                        trained_samples=trained_samples,
                        optimizer=optimizer)
                    better = "*"

                if time.time() - last_saved_time > Config.save_per_second:
                    save(global_steps = global_steps,
                         loss = total_loss,
                         prec = total_prec,
                         keep_name=True,
                         net = net,
                         trained_samples=trained_samples,
                         optimizer=optimizer)
                    last_saved_time = time.time()


                msg = 'Valid*  iter: {0:>5}, epoch: {1:>2},  word loss: {2:>6.2}, word prec: {3:>6.2%}, order loss: {4:>6.2}, order prec: {5:>6.2%},  ranking loss: {6:>6.2}, ranking prec: {7:>6.2%}{8}'
                logging.info(msg.format(global_steps, epoch, word_vloss, word_vprec, order_vloss, order_vprec, ranking_vloss, ranking_vprec, better))
                '''超过一定的batch验证集没变好，就early stop''' 
                #  if global_steps - last_saved_step > Config.early_stopping_batch:
                #      break
            
            # 达到lr_decay_batch 验证集没有变好，需要降低学习率
            if global_steps-last_saved_step > Config.lr_decay_batch and global_steps-last_adjust_lr_step > Config.lr_decay_batch:
               
                last_adjust_lr_step = global_steps
                msg = decay_lr(optimizer)
                logging.info(msg)

                
if __name__ == '__main__':
    
    train_folder = Folder(filepath=Config.train_path,  num_samples = None, is_test=False, num_umbc=0)
    train_loader = DataLoader(train_folder,
                              batch_size=Config.batch_size,
                              num_workers = 5,
                              pin_memory=False,
                              collate_fn=_collate_fn,
                              shuffle=True,
                              drop_last=True)
    
    valid_folder = Folder(filepath=Config.valid_path, num_samples = 5000, is_test=False, num_umbc=0)
    valid_loader = DataLoader(valid_folder, 
                              batch_size = Config.batch_size, 
                              num_workers = 5,
                              pin_memory=False,
                              collate_fn=_collate_fn,
                              shuffle = True,
                              drop_last=True)
    
    gpus = [_ for _ in args.gpu.split('-')]
    os.environ["CUDA_VISIBLE_DEVICES"] = ','.join(gpus)
    device = torch.device('cuda')
    net = SeqEncoder(Config)
    checkpoint = None
    #checkpoint = torch.load('./model/pretrained.w2v/GRU.case/word-order.maxpooling/best.pkl')
    train(train_loader = train_loader,
          valid_loader = valid_loader,
          checkpoint = checkpoint,
          net = net)
