# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Compute an expanded vocabulary of embeddings using a word2vec model.

This script loads the word embeddings from a trained skip-thoughts model and
from a trained word2vec model (typically with a larger vocabulary). It trains a
linear regression model without regularization to learn a linear mapping from
the word2vec embedding space to the skip-thoughts embedding space. The model is
then applied to all words in the word2vec vocabulary, yielding vectors in the
skip-thoughts word embedding space for the union of the two vocabularies.

The linear regression task is to learn a parameter matrix W to minimize
  || X - Y * W ||^2,
where X is a matrix of skip-thoughts embeddings of shape [num_words, dim1],
Y is a matrix of word2vec embeddings of shape [num_words, dim2], and W is a
matrix of shape [dim2, dim1].

This is based on the "Translation Matrix" method from the paper:

  "Exploiting Similarities among Languages for Machine Translation"
  Tomas Mikolov, Quoc V. Le, Ilya Sutskever
  https://arxiv.org/abs/1309.4168
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import os.path
import io

import gensim.models
import numpy as np
import sklearn.linear_model
import tensorflow as tf
import torch
import torch as nn
import argparse 


from net import *
from folder import *
from config import Config

tf.logging.set_verbosity(tf.logging.INFO)

parser = argparse.ArgumentParser()
parser.add_argument("-mp", "--model_path", help="pretrained model path")
parser.add_argument("-voc", "--voc_path", help="model vocab")
parser.add_argument("-glove", "--glove_path", help="pretrained glove path")
parser.add_argument("-out", "--output_dir", help="dir to save the result")
parser.add_argument("-gpu", "--gpu",  default='0', help="which gpu to run")

args = parser.parse_args()

def _load_sentEnc_embeddings(checkpoint_path):

    """Loads the embedding matrix from a sentence encoder model checkpoint.

    Args:
    checkpoint_path: Model checkpoint file or directory containing a checkpoint
        file.

    Returns:
    word_embedding: A numpy array of shape [vocab_size, embedding_dim].

    Raises:
    ValueError: If no checkpoint file matches checkpoint_path.
    """

    net = SeqEncoder(Config)
    checkpoint = torch.load(checkpoint_path)
    if 'net' in checkpoint:
        net.load_state_dict(checkpoint['net'])
    else:
        net.load_state_dict(checkpoint)

    word_embedding = net.embedding.weight.data.numpy()

    tf.logging.info("Loaded sentence encoder embedding matrix of shape %s", word_embedding.shape)

    return word_embedding


def _load_vocabulary(filename):
    """Loads a vocabulary file.

    Args:
    filename: Path to text file containing newline-separated words.

    Returns:
    vocab: A dictionary mapping word to word id.
    """
    vocab = json.load(open(filename))
    tf.logging.info("Read vocabulary of size %d", len(vocab))
    return vocab

# Get word vectors from vocabulary (glove, word2vec, fasttext ..)
def get_wordvec(path_to_vec):     
    
    word_vec = {}      
    with io.open(path_to_vec, 'r', encoding='utf-8') as f:
        # if word2vec or fasttext file : skip first line "next(f)"
        for line in f:
            word, vec = line.split(' ', 1)
            word_vec[word] = np.fromstring(vec, sep=' ')      
    
    return word_vec 


def _expand_vocabulary(sentEnc_emb, model_vocab, word2vec):
    """Runs vocabulary expansion on a sentence encoder model using a word2vec model.

    Args:
    sentEnc_emb: A numpy array of shape [model_vocab_size,
        sentEnc_embedding_dim].
    model_vocab: A dictionary of word to id.
    word2vec: An instance of gensim.models.Word2Vec.

    Returns:
    combined_emb: A dictionary mapping words to embedding vectors.
    """
    # Find words shared between the two vocabularies.
    tf.logging.info("Finding shared words")
    shared_words = [w for w in word2vec.keys() if w in model_vocab]

    # Select embedding vectors for shared words.
    tf.logging.info("Selecting embeddings for %d shared words", len(shared_words))
    shared_st_emb = sentEnc_emb[[
      model_vocab[w] for w in shared_words
    ]]
    shared_w2v_emb = [word2vec[w] for w in shared_words]

    # Train a linear regression model on the shared embedding vectors.
    tf.logging.info("Training linear regression model")
    model = sklearn.linear_model.LinearRegression()
    model.fit(shared_w2v_emb, shared_st_emb)

    # Create the expanded vocabulary.
    tf.logging.info("Creating embeddings for expanded vocabuary")
    combined_emb = collections.OrderedDict()
    for w in word2vec:
    # Ignore words with underscores (spaces).
        if "_" not in w:
            w_emb = model.predict(word2vec[w].reshape(1, -1))
            combined_emb[w] = w_emb.reshape(-1)

    for w in model_vocab:
        combined_emb[w] = sentEnc_emb[model_vocab[w]]

    voc = {}
    for i, w in enumerate(combined_emb.keys()):
        voc[w] = i

    tf.logging.info("Created expanded vocabulary of %d words", len(combined_emb))

    return combined_emb, voc


def main():
    
    # Load the sentEnc embeddings and vocabulary.
    
    sentEnc_emb = _load_sentEnc_embeddings(args.model_path)
    
    model_vocab = _load_vocabulary(args.voc_path)
    
    # Load the Word2Vec model.
    word2vec = get_wordvec(args.glove_path)

    # Run vocabulary expansion.
    embedding_map, vocab = _expand_vocabulary(sentEnc_emb, model_vocab, word2vec)

    # Save the output.
    vocab_file = os.path.join(args.output_dir, 'expand_voc.json')
    json.dump(vocab, open(vocab_file, 'w'))
    tf.logging.info("Wrote vocabulary file to %s", vocab_file)

    embeddings = np.array(list(embedding_map.values()))
    embeddings_file = os.path.join(args.output_dir, 'expand_vec.npy')
    np.save(embeddings_file, embeddings)
    tf.logging.info("Wrote embeddings file to %s", embeddings_file)


if __name__ == "__main__":
    
    os.environ['CUDA_VISIBLE_DEVICES'] = ','.join(args.gpu.split('-'))
    main()
